#include "gif_lib.h"
#include <stdio.h>
#include <string.h>
#include <vector>
#include <string>

#include <getopt.h>


struct ImageRecord
{
    int delay; // delay after showing this pic
    unsigned char DelayExt[4];
    
    GifImageDesc desc;
    ColorMapObject cmap;
    std::vector<GifPixelType> pixels;
};
struct DataRecord
{
    int type;
    std::vector<std::vector<unsigned char> > records;
};
struct OutputRecord : public ImageRecord, public DataRecord
{
    enum { t_img, t_data } type;

    OutputRecord() { }
    OutputRecord(const ImageRecord& d) : ImageRecord(d), type(t_img) { }
    OutputRecord(const DataRecord& d) : DataRecord(d), type(t_data) { }
};
struct OutputData : private std::vector<OutputRecord>
{
    using std::vector<OutputRecord>::operator[];
    using std::vector<OutputRecord>::size;
    
    int LastImg;
    
    OutputData() : LastImg(-1) { }
    
    void push_back(const OutputRecord& d)
    {
        if(d.type == OutputRecord::t_img) LastImg = size();
        std::vector<OutputRecord>::push_back(d);
    }
};

int main(int argc, char** argv)
{
    std::string infn, outfn;
    
    for(;;)
    {
        int option_index = 0;
        static struct option long_options[] =
        {
            {"help",     0, 0,'h'},
            {"version",  0, 0,'V'},
            {0,0,0,0}
        };
        int c = getopt_long(argc, argv, "hV", long_options, &option_index);
        if(c==-1) break;
        switch(c)
        {
            case 'V':
            {
                printf("%s\n", VERSION);
                return 0;
            }
            case 'h':
            {
                printf("%s", 
                    "gifshrenk v"VERSION" - Copyright (C) 1992,2006 Bisqwit (http://iki.fi/bisqwit/)\n"
                    "\n"
                    "Usage: gifshrenk [<options>] <inputfile> <outputfile>\n"
                    "\n"
                    "Removes completely transparent frames from a GIF file.\n"
                    " --help, -h     This help\n"
                    " --version, -V  Displays version information\n"
                    "\n");
                return 0;
            }
            default: goto ArgError;
        }
    }
    
    if(argc != optind+2)
    {
    ArgError:
        fprintf(stderr, "gifshrenk: Invalid parameters. Try `gifshrenk --help'\n");
        return 1;
    }
    
    infn  = argv[optind+0];
    outfn = argv[optind+1];
    
    GifFileType* i = DGifOpenFileName(infn.c_str());
    if(!i)
    {
        perror(infn.c_str());
        return -1;
    }
    GifFileType* o = EGifOpenFileName(outfn.c_str(), 0);
    if(!o)
    {
        perror(outfn.c_str());
        DGifCloseFile(i);
        return -1;
    }

    EGifPutScreenDesc(o,
        i->SWidth, i->SHeight,
        i->SColorResolution,
        i->SBackGroundColor,
        i->SColorMap);

    OutputData result;
    
    unsigned char DelayExt[4];
    int delay       = 0;
    int transparent = -1;
    
    for(;;)
    {
        GifRecordType RecordType;
        if(DGifGetRecordType(i, &RecordType) == GIF_ERROR) break;
        switch(RecordType)
        {
            case IMAGE_DESC_RECORD_TYPE:
            {
                if(DGifGetImageDesc(i) != GIF_ERROR)
                {
                    bool empty = true;
                    
                    const unsigned x = i->Image.Width;
                    const unsigned y = i->Image.Height;
                    
                    GifPixelType image[x * y];
                    
                    for(unsigned a=0; a<y; ++a)
                        DGifGetLine(i, image+a*x, x);
                        
                    for(unsigned a=0; a<x*y; ++a)
                        if(image[a] != transparent) { empty = false; break; }
                    
                    ImageRecord rec;
                    rec.desc = i->Image;
                    if(i->Image.ColorMap)
                        rec.cmap = *i->Image.ColorMap;
                    else
                        rec.cmap.ColorCount = 0;
                    rec.pixels.assign(image, image+x*y);
                    rec.delay = 0;
                    
                    if(empty) break;

                    std::memcpy(rec.DelayExt, DelayExt, 4);
                    
                    rec.delay += delay;
                    delay = 0;
                        
                    result.push_back(rec);
                }
                break;
            }
            case EXTENSION_RECORD_TYPE:
            {
                int ExtCode;
                GifByteType* Extension;
                if(DGifGetExtension(i, &ExtCode, &Extension) == GIF_ERROR) break;
                
                std::vector<std::vector<unsigned char> > ext_records;
                
                bool Ignore = false;
                bool NetscapeExt = false;
                switch(ExtCode)
                {
                    case COMMENT_EXT_FUNC_CODE: break;
                    case GRAPHICS_EXT_FUNC_CODE:
                    {
                        memcpy(DelayExt, Extension+1, sizeof(DelayExt));
                        
                        unsigned newdelay = (Extension[2] * 1) | (Extension[3] * 256);
                        delay += newdelay;
                        transparent = (Extension[1] & 1) ? Extension[4] : -1;
                        if(result.LastImg >= 0)
                        {
                            result[result.LastImg].delay += delay;
                            delay = 0;
                        }
                        Ignore = true;
                        break;
                    }
                    case PLAINTEXT_EXT_FUNC_CODE: break;
                    case APPLICATION_EXT_FUNC_CODE:
                        NetscapeExt = strncmp((const char*)&*Extension, "NETSCAPE2.0", 12) == 0;
                        break;
                    default: break;
                }
                
                while(Extension)
                {
                    ext_records.push_back(std::vector<unsigned char> (Extension+1, Extension+1+*Extension) );
                    
                    if(DGifGetExtensionNext(i, &Extension) == GIF_ERROR) break;
                    if(!Extension) break;
                }
                
                if(!Ignore)
                {
                    DataRecord rec;
                    rec.type = ExtCode;
                    rec.records = ext_records;
                    result.push_back(rec);
                }
                break;
            }
            case UNDEFINED_RECORD_TYPE: // ignore
                break;
            case SCREEN_DESC_RECORD_TYPE: // ignore
                break;
            case TERMINATE_RECORD_TYPE: // ignore
                break;
        }
    }
    
    DGifCloseFile(i);
    
    for(unsigned a=0; a<result.size(); ++a)
        switch(result[a].type)
        {
            case OutputRecord::t_img:
            {
                ImageRecord& rec = result[a];
                std::vector<GifPixelType>& image = rec.pixels;
                
                if(rec.delay > 0)
                {
                    rec.DelayExt[1] = rec.delay & 0xFF;
                    rec.DelayExt[2] = rec.delay >> 8;
                    
                    if(rec.delay > 65535)
                    {
                        fprintf(stderr, "Warning: Too great delay (%.2f seconds), saved as %.2f seconds\n",
                            (rec.delay) / 100.0,
                            (rec.delay & 0xFFFF) / 100.0
                               );
                    }
                    EGifPutExtension(o, GRAPHICS_EXT_FUNC_CODE, 4, rec.DelayExt);
                }
                    
                const unsigned x = rec.desc.Width;
                const unsigned y = rec.desc.Height;
                
                EGifPutImageDesc(o, rec.desc.Left, rec.desc.Top,
                                    x, y,
                                    rec.desc.Interlace,
                                    rec.cmap.ColorCount ? &rec.cmap : NULL);
                for(unsigned a=0; a<y; ++a)
                    EGifPutLine(o, &image[a*x], x);
                break;
            }
            case OutputRecord::t_data:
            {
                DataRecord& rec = result[a];
                int ExtCode = rec.type;
                
                for(unsigned a=0; a<rec.records.size(); ++a)
                {
                    const std::vector<unsigned char>& record = rec.records[a];
                    if(a == 0)
                        if(rec.records.size() > 1)
                            EGifPutExtensionFirst(o, ExtCode, record.size(), &record[0]);
                        else
                            EGifPutExtension(o, ExtCode, record.size(), &record[0]);
                    else if(a+1 == rec.records.size())
                        EGifPutExtensionLast(o, ExtCode, record.size(), &record[0]);
                    else
                        EGifPutExtensionNext(o, ExtCode, record.size(), &record[0]);
                }
                break;
            }
        }

    EGifCloseFile(o);
    
    return 0;
}
